// -*- C++ -*-
/*!
 * @file RtcBase.cpp
 * @brief RT component base class
 * @date $Date$
 *
 * Copyright (c) 2003,2004 Noriaki Ando <n-ando@aist.go.jp>
 *          Task-intelligence Research Group,
 *          Intelligent System Research Institute,
 *          National Institute of Industrial Science (AIST), Japan
 *          All rights reserved.
 *
 * $Id$
 */

#include "rtm/RtcBase.h"
#include "rtm/RtcInPortBase.h"
#include "rtm/RtcOutPortBase.h"
#include "rtm/RtcManager.h"
#include <ace/OS.h>

namespace RTM {

  RtcBase::RtcBase()
  	: m_StatePort("status", m_TimedState),
	  m_NamingPolicy((NamingPolicy)(LONGNAME_ENABLE | ALIAS_ENABLE))
  {
	ACE_TRACE("RtcBase::RtcBase()");

	init_state_func_table();
  }

  RtcBase::RtcBase(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa)
  	: m_StatePort("status", m_TimedState),
	  m_NamingPolicy((NamingPolicy)(LONGNAME_ENABLE | ALIAS_ENABLE)),
	  m_pORB(orb), m_pPOA(poa)
  {
	ACE_TRACE("RtcBase::RtcBase(CORBA::ORB_var, PortableServer::POA_var)");

	init_state_func_table();
  }

  RtcBase::RtcBase(RtcManager* manager)
	: m_StatePort("status", m_TimedState),
	  m_NamingPolicy((NamingPolicy)(LONGNAME_ENABLE | ALIAS_ENABLE)),
	  m_pORB(manager->getORB()), m_pPOA(manager->getPOA())
  {
	ACE_TRACE("RtcBase::RtcBase(RtcManager*)");
	/*
	CORBA::PolicyList policies;
	policies.length(0);
	PortableServer::POA_var rootPOA = manager->getPOA();
	PortableServer::POAManager_var rootManager = rootPOA->the_POAManager(); 
	m_pPOA = rootPOA->create_POA("hoge",
								 rootManager, policies);
	//	m_pPOA->the_POAManager()->activate();
	*/
	init_state_func_table();
  }


  RtcBase::~RtcBase()
  {
	ACE_TRACE("RtcBase::~RtcBase()");
	
  }


  //============================================================
  //
  // <<< CORBA interfaces >>>
  //
  //============================================================
  

  //------------------------------------------------------------
  // RTC_READY -> RTC_STARTING -> RTC_ACTIVE
  //
  // Activate this components' activity.
  // Other components or managers call this method to activate
  // the component. Component should be in Ready state, when
  // this method is called.
  //------------------------------------------------------------
  RtmRes RtcBase::rtc_start()
	throw (CORBA::SystemException, RTM::RTComponent::IllegalTransition)
  {
	ACE_TRACE("RtcBase::rtc_start()");

	ACE_Guard<ACE_Thread_Mutex> guard_curr(m_CurrentState._mutex);

	if (m_CurrentState._state == RTM::RTComponent::RTC_READY)
	  {
		ACE_Guard<ACE_Thread_Mutex> guard_next(m_NextState._mutex);
		m_NextState._state = RTM::RTComponent::RTC_STARTING;
		return RTM_OK;
	  }
	else
	  {
		throw RTM::RTComponent::IllegalTransition();
	  }
	return RTM_ERR;
  }


  //------------------------------------------------------------
  // RTC_ACTIVE -> RTC_STOPPING -> RTC_ABORTING
  //
  // Stop this component's activity.
  // Other components or managers call this method to stop the
  // component. Component should be in Active state when this
  // method is called.
  //------------------------------------------------------------
  RtmRes RtcBase::rtc_stop()
	  throw (CORBA::SystemException, RTM::RTComponent::IllegalTransition)
  {
	ACE_TRACE("RtcBase::rtc_stop()");

	ACE_Guard<ACE_Thread_Mutex> guard_curr(m_CurrentState._mutex);

	if (m_CurrentState._state == RTM::RTComponent::RTC_ACTIVE)
	  {
		ACE_Guard<ACE_Thread_Mutex> guard_next(m_NextState._mutex);
		m_NextState._state = RTM::RTComponent::RTC_STOPPING;
		return RTM_OK;
	  }
	else
	  {
		throw RTM::RTComponent::IllegalTransition();
	  }
	return RTM_ERR;
  }


  //------------------------------------------------------------
  // RTC_ERROR -> RTC_INITIALIZING -> RTC_READY
  // RTC_READY -> RTC_INITIALIZING -> RTC_READY
  //
  // Reset this component to recover from error state.
  // Other component or managers call this method to recover
  // the component from error. After transit initializing state
  // component's state will go to Ready state.
  //------------------------------------------------------------
  RtmRes RtcBase::rtc_reset()
	  throw (CORBA::SystemException, RTM::RTComponent::IllegalTransition)
  {
	ACE_TRACE("RtcBase::rtc_reset()");

	ACE_Guard<ACE_Thread_Mutex> guard_curr(m_CurrentState._mutex);

	if (m_CurrentState._state == RTM::RTComponent::RTC_READY ||
		m_CurrentState._state == RTM::RTComponent::RTC_ERROR)
	  {
		ACE_Guard<ACE_Thread_Mutex> guard_next(m_NextState._mutex);
		m_NextState._state = RTM::RTComponent::RTC_INITIALIZING;
		return RTM_OK;
	  }
	else
	  {
		throw RTM::RTComponent::IllegalTransition();
	  }
	return RTM_ERR;
  }
  
  
  //------------------------------------------------------------
  // RTC_READY -> RTC_EXITING -> END
  // RTC_ERROR -> RTC_EXITING -> END
  //
  // Terminate this component to delete this component.
  // Other components or managers can call this method when 
  // the component is in Error state or Ready state to terminate
  // the component.
  //------------------------------------------------------------
  RtmRes RtcBase::rtc_exit()
	  throw (CORBA::SystemException, RTM::RTComponent::IllegalTransition)
  {
	ACE_TRACE("RtcBase::rtc_exit()");

	ACE_Guard<ACE_Thread_Mutex> guard_curr(m_CurrentState._mutex);
	
	if (m_CurrentState._state == RTM::RTComponent::RTC_READY ||
		m_CurrentState._state == RTM::RTComponent::RTC_ERROR)
	  {
		ACE_Guard<ACE_Thread_Mutex> guard_next(m_NextState._mutex);
		m_NextState._state = RTM::RTComponent::RTC_EXITING;
		return RTM_OK;
	  }
	else
	  {
		throw RTM::RTComponent::IllegalTransition();
	  }
	return RTM_ERR;
  }
  

  //------------------------------------------------------------
  // -> RTC_FATAL_ERROR -> END
  // resume: Go to error() state from anywhrer
  //------------------------------------------------------------
  RtmRes RtcBase::rtc_kill()
  {
	ACE_TRACE("RtcBase::rtc_kill()");

	ACE_Guard<ACE_Thread_Mutex> guard_curr(m_CurrentState._mutex);
	
	if (m_CurrentState._state == RTM::RTComponent::RTC_FATAL_ERROR)
	  {
		ACE_Guard<ACE_Thread_Mutex> guard_next(m_NextState._mutex);
		m_NextState._state = RTM::RTComponent::RTC_EXITING;
		return RTM_OK;
	  }
	return RTM_ERR;
  }
  


  //------------------------------------------------------------
  // rtc_main: Main activity function. Called from activity thread periodically.
  //------------------------------------------------------------
  RtmRes RtcBase::rtc_worker()
  {
	ACE_TRACE("RtcBase::rtc_worker()");

	ComponentState cur_state;
	ComponentState nxt_state;
	{
	  ACE_Guard<ACE_Thread_Mutex> guard_curr(m_CurrentState._mutex);
	  ACE_Guard<ACE_Thread_Mutex> guard_next(m_NextState._mutex);
	  cur_state = m_CurrentState._state;
	  nxt_state = m_NextState._state;
	  m_CurrentState._state = m_NextState._state;
	  m_TimedState.data = m_NextState._state;
	  m_StatePort.write();
	}
	// Change state
	//	ACE_DEBUG((LM_DEBUG,"Current stat: %s, Next stat: %s\n",
	//			   StateString[m_CurrentState._state],
	//			   StateString[m_NextState._state]));
	// Status was changed
	if (cur_state != nxt_state)
	  {
		// exit from the current state.
		RtmRes result_exit = (this->*_exit_func[cur_state])();
		_check_error(result_exit);
		
		// Enter to the next state.
		// If next state is transient state, execute transient state function.
		RtmRes result_entry = (this->*_entry_func[nxt_state])();
		return _check_error(result_entry);
	  }
	else 
	  {
		// ֤ϤΤޤޤʤΤdo ¹
		RtmRes result_do = (this->*_do_func[cur_state])();
		return _check_error(result_do);
	  }
	
	return RTM_ERR;
  }


  //------------------------------------------------------------
  // _check_error: Error check help function
  //------------------------------------------------------------
  RtmRes RtcBase::_check_error(RtmRes result)
  {
	ACE_TRACE("RtcBase::_check_error(RtmRes result)");

	// If current function returns error, goto RTC_ERROR status.
	if (result == RTM_ERR && m_CurrentState._state
		!= RTM::RTComponent::RTC_ACTIVE) 
	  {
		m_NextState._state = RTM::RTComponent::RTC_ERROR;

		return RTM_ERR;
	  }
	//
	// Active state have to transit RTC_ABORTING status.
	//
	else if (result == RTM_ERR
			 && m_CurrentState._state == RTM::RTComponent::RTC_ACTIVE) 
	  {
		m_NextState._state = RTM::RTComponent::RTC_ABORTING;

		return RTM_ERR;
	  }
	//
	// If function returns RTM_FATAL_ERR, status is directly changed
	//  RTC_FATAL_ERRROR
	else if (result == RTM_FATAL_ERR)
	  {
		m_NextState._state = RTM::RTComponent::RTC_FATAL_ERROR;

		return RTM_FATAL_ERR;
	  }

	return RTM_ERR;
  }


  //------------------------------------------------------------
  // Status action functions
  //
  // This functions should be only called rtc_worker.
  // m_CurrentState and m_NextState are already locked.
  // So these variables need not lock once again.
  //------------------------------------------------------------  
  RtmRes RtcBase::_rtc_initializing()
  {
	ACE_TRACE("RtcBase::_rtc_initializing()");

	m_NextState._state = RTM::RTComponent::RTC_READY;

	return this->rtc_init_entry();
  }

  RtmRes RtcBase::_rtc_ready_entry()
  {
	ACE_TRACE("RtcBase::_rtc_ready_entry()");

	return this->rtc_ready_entry();
  }
  
  RtmRes RtcBase::_rtc_starting()
  {
	ACE_TRACE("RtcBase::_rtc_starting()");

	m_NextState._state = RTM::RTComponent::RTC_ACTIVE;

	return this->rtc_starting_entry();
  }

  RtmRes RtcBase::_rtc_active_entry()
  {
	ACE_TRACE("RtcBase::_rtc_active_entry()");

	return this->rtc_active_entry();
  }

  RtmRes RtcBase::_rtc_stopping()
  {
	ACE_TRACE("RtcBase::_rtc_stopping()");

	m_NextState._state = RTM::RTComponent::RTC_READY;

	return this->rtc_stopping_entry();
  }

  RtmRes RtcBase::_rtc_aborting()
  {
	ACE_TRACE("RtcBase::_rtc_aborting()");

	m_NextState._state = RTM::RTComponent::RTC_ERROR;

	return this->rtc_aborting_entry();
  }

  RtmRes RtcBase::_rtc_error_entry()
  {
	ACE_TRACE("RtcBase::_rtc_error_entry()");

	return this->rtc_error_entry();
  }

  RtmRes RtcBase::_rtc_fatal_entry()
  {
	ACE_TRACE("RtcBase::_rtc_fatal_entry()");

	return this->rtc_fatal_entry();
  }

  RtmRes RtcBase::_rtc_exiting()
  {
	ACE_TRACE("RtcBase::_rtc_exiting()");

	m_NextState._state = RTM::RTComponent::RTC_UNKNOWN;

	return this->rtc_exiting_entry();
  }

  RtmRes RtcBase::_rtc_ready()
  {
	return this->rtc_ready_do();
  }

  RtmRes RtcBase::_rtc_active()
  {
	return this->rtc_active_do();
  }

  RtmRes RtcBase::_rtc_error()
  {
	return this->rtc_error_do();
  }

  RtmRes RtcBase::_rtc_fatal()
  {
	return this->rtc_fatal_do();
  }

  RtmRes RtcBase::_rtc_ready_exit()
  {
	ACE_TRACE("RtcBase::_rtc_ready_exit()");

	return this->rtc_ready_exit();
  }
  
  RtmRes RtcBase::_rtc_active_exit()
  {
	ACE_TRACE("RtcBase::_rtc_active_exit()");

	return this->rtc_active_exit();
  }
  
  RtmRes RtcBase::_rtc_error_exit()
  {
	ACE_TRACE("RtcBase::_rtc_error_exit()");

	m_NextState._state = RTM::RTComponent::RTC_EXITING;
	return this->rtc_error_exit();
  }
  
  RtmRes RtcBase::_rtc_fatal_exit()
  {
	ACE_TRACE("RtcBase::_rtc_fatal_exit()");

	return this->rtc_fatal_exit();
  }
  // End of status fanctions.
  //------------------------------------------------------------





  //------------------------------------------------------------
  // rtc_state()
  //------------------------------------------------------------
  OutPort_ptr RtcBase::rtc_state()
  {
	ACE_TRACE("RtcBase::rtc_state()");

	return OutPort::_narrow(m_pPOA->servant_to_reference(&m_StatePort));
  }
  
  
  InPortList* RtcBase::inports()
  {
	ACE_TRACE("RtcBase::inports()");

	ACE_Guard<ACE_Thread_Mutex> guard(m_InPorts.m_Mutex);
	
	InPortList_var inport_list = new InPortList();
	inport_list->length(m_InPorts.m_List.size());
	
	InPorts_it it = m_InPorts.m_List.begin();
	InPorts_it it_end = m_InPorts.m_List.end();
	int i(0);
	
	while (it != it_end)
	  {
		//
		// POA should have policies RETAIN & UNIQUE_ID. 
		// If POA have MULTIPLE_ID policy, InPort newly activate on the POA.
		// 
		inport_list[i] = InPort::_narrow(m_pPOA->servant_to_reference(*it));
		it++;
		i++;
	}
	
	return inport_list._retn();
  }
  
  
  InPort_ptr RtcBase::get_inport(const char* name)
	throw (CORBA::SystemException, RTM::RTComponent::NoSuchName)
  {
	ACE_TRACE("RtcBase::get_inport()");
	
	ACE_Guard<ACE_Thread_Mutex> guard(m_InPorts.m_Mutex);
	
	InPorts_it it;
	it = find_if(m_InPorts.m_List.begin(),
				 m_InPorts.m_List.end(),
				 eq_name(name));
	InPort_var inp = InPort::_narrow(m_pPOA->servant_to_reference(*it));
	if (it == m_InPorts.m_List.end())
	  {
		throw RTM::RTComponent::NoSuchName(name);
	  }
	return inp._retn();
  }
  
  
  OutPortList* RtcBase::outports()
  {
	ACE_TRACE("RtcBase::outports()");

	ACE_Guard<ACE_Thread_Mutex> guard(m_OutPorts.m_Mutex);
	
	OutPortList_var outport_list = new OutPortList();
	outport_list->length(m_OutPorts.m_List.size());

	OutPorts_it it = m_OutPorts.m_List.begin();
	OutPorts_it it_end = m_OutPorts.m_List.end();
	int i(0);

	while (it != it_end)
	  {	
		//
		// POA should have policies RETAIN & UNIQUE_ID. 
		// If POA have MULTIPLE_ID policy, InPort newly activate on the POA.
		//
		outport_list[i] = OutPort::_narrow(m_pPOA->servant_to_reference(*it));
		it++;
		i++;
	  }
	
	return outport_list._retn();
  }
  
  
  OutPort_ptr RtcBase::get_outport(const char* name)
	throw (CORBA::SystemException, RTM::RTComponent::NoSuchName)
	
  {
	ACE_TRACE("RtcBase::get_outport()");

	ACE_Guard<ACE_Thread_Mutex> guard(m_OutPorts.m_Mutex);
	
	OutPorts_it it;
	it = find_if(m_OutPorts.m_List.begin(),
				 m_OutPorts.m_List.end(),
				 eq_name(name));
	
	OutPort_var outp = OutPort::_narrow(m_pPOA->servant_to_reference(*it));
	if (it == m_OutPorts.m_List.end())
	  {
		throw RTM::RTComponent::NoSuchName(name);
	  }
	return outp._retn();
  }
  
  
  char* RtcBase::instance_id()
  {
	return CORBA::string_dup(m_Profile.getInstanceId());
  }

  char* RtcBase::implementation_id()
  {
	return CORBA::string_dup(m_Profile.getImplementationId());
  }

  char* RtcBase::description()
  {
	return CORBA::string_dup(m_Profile.getDescription());
  }

  char* RtcBase::version()
  {
	return CORBA::string_dup(m_Profile.getVersion());
  }
  
  char* RtcBase::maker()
  {
	return CORBA::string_dup(m_Profile.getAuthorName());
  }

  char* RtcBase::category()
  {
	return CORBA::string_dup(m_Profile.getCategory());
  }


  
  
  RTCProfile* RtcBase::profile()
  {
	ACE_TRACE("RtcBase::profile()");

	RTCProfile_var profile = new RTCProfile(m_Profile.getProfile());
	return profile._retn();
  }

  /*
  RTCConfiguration* RtcBase::configuration()
  {
	ACE_TRACE("RtcBase::configuration()");

	RTCConfiguration_var config = 
	  new RTCConfiguration(m_Configuration.getConfiguration());
	return config._retn();
  }
  */
  RtmRes RtcBase::rtc_start_thread()
  {
	//	ACE_Guard<ACE_Thread_Mutex> guard_next(m_ThreadState.m_Mutex);
	m_ThreadState.m_Flag = true;
	open(0);
	return RTM_OK;
  }


  RtmRes RtcBase::rtc_stop_thread()
  {
	std::cout << "RtcBase::rtc_stop_thread() " << std::endl;
  
	//	ACE_Guard<ACE_Thread_Mutex> guard_next(m_ThreadState.m_Mutex);
	m_ThreadState.m_Flag = false;
	//	close(0);
	//	wait();
	return RTM_OK;
  }

  
  RtmRes RtcBase::rtc_set_parent(RTCBase_ptr comp)
  {
	if (CORBA::is_nil(comp))
	  {
		return RTM_ERR;
	  }
	m_Parent = RTM::RTCBase::_duplicate(comp);
	
	return RTM_OK;
  }
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  //============================================================
  // Local interfaces
  //============================================================
  void RtcBase::init_orb(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa)
  {
	m_pORB = orb;
	m_pPOA = poa;
  }
  
  /*------------------------------------------------------------
   * Start activity
   * ACE_Task class method over ride.
   *------------------------------------------------------------*/
  int RtcBase::open(void *args)
  {
	ACE_TRACE("RtcBase::open()");

	ACE_Guard<ACE_Thread_Mutex> guard_next(m_NextState._mutex);
	m_NextState._state = RTM::RTComponent::RTC_INITIALIZING;
	activate();
	return 0;
  }
  
  
  /*------------------------------------------------------------
   * Run by a daemon thread to handle deferred processing
   * ACE_Task class method over ride.
   *------------------------------------------------------------*/
  int RtcBase::svc(void)
  {
	ACE_TRACE("RtcBase::svc()");

	while (m_ThreadState.m_Flag == true)
	  {
		rtc_worker();
	  }
	return 0;
  }


  void RtcBase::initModuleProfile(RtcModuleProfile prof)
  {
	m_Profile.setInstanceId      (prof.getInstanceId());
	m_Profile.setImplementationId(prof.getImplementationId());
	m_Profile.setDescription     (prof.getDescription());
	m_Profile.setVersion         (prof.getVersion());
	m_Profile.setAuthorName      (prof.getAuthorName());
	m_Profile.setCategory        (prof.getCategory());
	m_Profile.setComponentType   (prof.getComponentType());
	m_Profile.setActivityType    (prof.getActivityType());
	m_Profile.setMaxInstance     (prof.getMaxInstance());
	m_Profile.setLanguage        (prof.getLanguage());
	m_Profile.setLanguageType    (prof.getLanguageType());
	m_Profile.setIconImage       (prof.getIconImage());
  }

  
  /*------------------------------------------------------------
   * Set component name
   *------------------------------------------------------------*/  
  string RtcBase::setComponentName(int num)
  {
	string module_name(m_Profile.getImplementationId());
	string comp_name;
	char n[1000];
	
	ACE_OS::itoa(num, n, 10);
	string nn(n);
	
	comp_name = module_name + nn;

	m_Profile.setInstanceId(comp_name.c_str());

	return comp_name;
  }


 
  /*------------------------------------------------------------
   * Protcted member function for Component Creator
   *------------------------------------------------------------*/
  bool RtcBase::registerInPort(InPortBase& inport)
  {
	ACE_TRACE("RtcBase::registerInPort(InPortBase&)");

	ACE_Guard<ACE_Thread_Mutex> guard(m_InPorts.m_Mutex);
	
	// ǤƱ̾Υͥ뤬ʤɤå
	InPorts_it it;
	it = find_if(m_InPorts.m_List.begin(),
				 m_InPorts.m_List.end(),
				 eq_name(inport.name()));
	
	if (it != m_InPorts.m_List.end())
	  {
		return false;
	  }
	
	// Store input channel reference and flag.
	m_InPorts.m_List.push_back(&inport);
	//  m_inFlags.push_back(inport.getFlag());
	
	// append port profile to component profile
	m_Profile.appendInPortProfile(*(inport.profile()));

	// Activate Input Port servant
	m_pPOA->activate_object(&inport);
	
	return true;
  }
  
  
  bool RtcBase::deleteInPort(InPortBase& inport)
  {
	ACE_Guard<ACE_Thread_Mutex> guard(m_InPorts.m_Mutex);
	
	return deleteInPortByName(inport.name());
  }
  
  
  bool RtcBase::deleteInPortByName(const char* name)
  {
	ACE_Guard<ACE_Thread_Mutex> guard(m_InPorts.m_Mutex);
	
	InPorts_it it;
	
	it = find_if(m_InPorts.m_List.begin(),
				 m_InPorts.m_List.end(),
				 eq_name(name));
	
	if (it == m_InPorts.m_List.end()) // Not found
	  {
		return false;
	  }
	else
	  {
		m_pPOA->deactivate_object(*m_pPOA->servant_to_id(*it));
		m_InPorts.m_List.erase(it);
		//	  m_InFlags.remove((*it)->getFlag());
	  }
	return true;
  }
  
  
  bool RtcBase::registerOutPort(OutPortBase& outport)
  {
	ACE_TRACE("RtcBase::registerOutPort(OutPortBase&)");

	ACE_Guard<ACE_Thread_Mutex> guard(m_OutPorts.m_Mutex);
	
	// ǤƱ̾Υͥ뤬ʤɤå
	OutPorts_it it;
	it = find_if(m_OutPorts.m_List.begin(),
				 m_OutPorts.m_List.end(),
				 eq_name(outport.name()));
	
	if (it != m_OutPorts.m_List.end())
	  {
		ACE_DEBUG((LM_INFO, "Already registered port: %s.\n",
				   outport.name()));
		return false;
	  }

	// Store output channel reference and flag.
	m_OutPorts.m_List.push_back(&outport);
	//  m_OutFlags.push_back(out_ch.getFlag());
	ACE_DEBUG((LM_DEBUG, "Number of ports: %d\n",
			   m_OutPorts.m_List.size()));

	m_Profile.appendOutPortProfile(*(outport.profile()));

	// Activate Output Port servant
	m_pPOA->activate_object(&outport);

	return true;
  }
  
  
  bool RtcBase::deleteOutPort(OutPortBase& outport)
  {
	return deleteOutPortByName(outport.name());
  }
  
  
  bool RtcBase::deleteOutPortByName(const char* name)
  {
	ACE_Guard<ACE_Thread_Mutex> guard(m_OutPorts.m_Mutex);
	
	OutPorts_it it;
	
	it = find_if(m_OutPorts.m_List.begin(),
				 m_OutPorts.m_List.end(),
				 eq_name(name));
	
	if (it == m_OutPorts.m_List.end())
	  {
		return false;
	  }
	else
	  {
		m_pPOA->deactivate_object(*m_pPOA->servant_to_id(*it));
		m_OutPorts.m_List.erase(it);
		//	  m_outFlags.remove((*it)->getFlag());
	  }
	
	return true;
  }
  

  void RtcBase::appendAlias(const char* alias)
  {
	appendAlias(std::string(alias));
	return;
  }


  void RtcBase::appendAlias(std::string alias)
  {
	m_Alias.push_back(alias);
	return;
  }


  std::list<string> RtcBase::getAliases()
  {
	return m_Alias;
  }


  void RtcBase::setNamingPolicy(NamingPolicy policy)
  {
	m_NamingPolicy = policy;
  }

  NamingPolicy RtcBase::getNamingPolicy()
  {
	return m_NamingPolicy;
  }

  bool RtcBase::isLongNameEnable()
  {
	if (m_NamingPolicy & LONGNAME_ENABLE != 0)
	  {
		return true;
	  }
	return false;
  }

  bool RtcBase::isAliasEnable()
  {
	if (m_NamingPolicy & ALIAS_ENABLE != 0)
	  {
		return true;
	  }
	return false;
  }

  
  void RtcBase::init_state_func_table()
  {
	m_NextState._state = RTM::RTComponent::RTC_BORN;
	m_CurrentState._state = RTM::RTComponent::RTC_BORN;
	
	_entry_func[RTM::RTComponent::RTC_UNKNOWN]      = &RtcBase::_nop;
	_entry_func[RTM::RTComponent::RTC_BORN]         = &RtcBase::_nop;
	_entry_func[RTM::RTComponent::RTC_INITIALIZING] = &RtcBase::_rtc_initializing;
	_entry_func[RTM::RTComponent::RTC_READY]        = &RtcBase::_rtc_ready_entry;
	_entry_func[RTM::RTComponent::RTC_STARTING]     = &RtcBase::_rtc_starting;
	_entry_func[RTM::RTComponent::RTC_ACTIVE]       = &RtcBase::_rtc_active_entry;
	_entry_func[RTM::RTComponent::RTC_STOPPING]     = &RtcBase::_rtc_stopping;
	_entry_func[RTM::RTComponent::RTC_ABORTING]     = &RtcBase::_rtc_aborting;
	_entry_func[RTM::RTComponent::RTC_ERROR]        = &RtcBase::_rtc_error_entry;
	_entry_func[RTM::RTComponent::RTC_FATAL_ERROR]  = &RtcBase::_rtc_fatal_entry;
	_entry_func[RTM::RTComponent::RTC_EXITING]      = &RtcBase::_rtc_exiting;
	
	_do_func[RTM::RTComponent::RTC_UNKNOWN]         = &RtcBase::_nop;
	_do_func[RTM::RTComponent::RTC_BORN]            = &RtcBase::_nop;
	_do_func[RTM::RTComponent::RTC_INITIALIZING]    = &RtcBase::_nop;
	_do_func[RTM::RTComponent::RTC_READY]           = &RtcBase::_rtc_ready;
	_do_func[RTM::RTComponent::RTC_STARTING]        = &RtcBase::_nop;
	_do_func[RTM::RTComponent::RTC_ACTIVE]          = &RtcBase::_rtc_active;
	_do_func[RTM::RTComponent::RTC_STOPPING]        = &RtcBase::_nop;
	_do_func[RTM::RTComponent::RTC_ABORTING]        = &RtcBase::_nop;
	_do_func[RTM::RTComponent::RTC_ERROR]           = &RtcBase::_rtc_error;
	_do_func[RTM::RTComponent::RTC_FATAL_ERROR]     = &RtcBase::_rtc_fatal;
	_do_func[RTM::RTComponent::RTC_EXITING]         = &RtcBase::_nop;
	
	_exit_func[RTM::RTComponent::RTC_UNKNOWN]       = &RtcBase::_nop;
	_exit_func[RTM::RTComponent::RTC_BORN]          = &RtcBase::_nop;
	_exit_func[RTM::RTComponent::RTC_INITIALIZING]  = &RtcBase::_nop;
	_exit_func[RTM::RTComponent::RTC_READY]         = &RtcBase::_rtc_ready_exit;
	_exit_func[RTM::RTComponent::RTC_STARTING]      = &RtcBase::_nop;
	_exit_func[RTM::RTComponent::RTC_ACTIVE]        = &RtcBase::_rtc_active_exit;
	_exit_func[RTM::RTComponent::RTC_STOPPING]      = &RtcBase::_nop;
	_exit_func[RTM::RTComponent::RTC_ABORTING]      = &RtcBase::_nop;
	_exit_func[RTM::RTComponent::RTC_ERROR]         = &RtcBase::_rtc_error_exit;
	_exit_func[RTM::RTComponent::RTC_FATAL_ERROR]   = &RtcBase::_rtc_fatal_exit;
	_exit_func[RTM::RTComponent::RTC_EXITING]       = &RtcBase::_nop;

	m_NextState._state = RTM::RTComponent::RTC_INITIALIZING;
  }

  
}; // end of namespace RTM
