// -*- C++ -*-
/*!
 * @file RtcManager.cpp
 * @brief RT component manager class
 * @date $Date$
 *
 * Copyright (c) 2003-2004 Noriaki Ando <n-ando@aist.go.jp>
 *          Task-intelligence Research Group,
 *          Intelligent System Research Institute,
 *          National Institute of Industrial Science (AIST),
 *          Japan. All rights reserved.
 *
 * $Id$
 */


#include <iostream>
#include <stdlib.h>
#include <boost/regex.hpp>

// ACE includes
#include <ace/Thread.h>
#include <ace/Trace.h>

// RTM includes
#include "rtm/RtcManager.h"
#include "rtm/RtcNaming.h"
#include "rtm/RtcConfig.h"


namespace RTM
{
  using namespace std;
  
  //------------------------------------------------------------
  // RtcManager(): Constructor
  //------------------------------------------------------------
  RtcManager::RtcManager(int argc, char** argv)
    : m_apConfig(new RtcConfig(argc, argv)),
      m_apNaming(new RtcNaming)
  {
    // ORB have to be initilialized  to use ORB function
    // ex. RtcNaming, POA, POAManager
    if (!initORB()) {
      cerr << "Error: initORB()" << endl;
	  exit(1);
    } 
	
    //! POA have to be initialized to use POA
    //! ex. activate RtcManager servant
    if (!initPOA()) {
      cerr << "Error: initPOA()" << endl;
	  exit(1);
    }
	ACE_Trace::start_tracing();
  }
  
  //------------------------------------------------------------
  // ~RtcManager(): destructor
  //------------------------------------------------------------
  RtcManager::~RtcManager()
  {
    ;
  }
  
  /*
  int RtcManager::open(void *args)
  {
	ACE_TRACE("RtcManager::open()");
	activate();
	return 0;;
  }
  */

  int RtcManager::svc(void)
  {
	runManager();
  }


  //============================================================
  //
  // <<< CORBA interfaces >>>
  //
  //============================================================
  
  /*!------------------------------------------------------------
   * [CORBA interface] load module
   * @brief Load shared library of a component by path name,
   *        and invoke initialize function.
   *------------------------------------------------------------*/
  RtmRes RtcManager::load(const char* libname, const char* initfunc)
  {
    string path(libname);
    string full_path;
	
	cout << "load()" << libname << ":" << initfunc << endl;

    if (!findPath(path, full_path))
      {
		cout << "path error: " << full_path << endl;
		// invalid libname
		return RTM_ERR;
      }
	
    string initfunc_name;
    libnameToInitFunc(libname, initfunc, initfunc_name);

	cout << "init func name" << initfunc_name << endl;
	
    // Open shared object (or DLL)
    int retval =  m_Lib[full_path]._dll.open(full_path.c_str());
	
    if (retval != 0)
      ACE_ERROR_RETURN ((LM_ERROR, "%p", "dll.open"), RTM_ERR);

    RtcComponentInit init_func;
    init_func = (RtcComponentInit) m_Lib[full_path]._dll.symbol(initfunc_name.c_str());
    if (init_func == NULL)
      ACE_ERROR_RETURN ((LM_ERROR, "%p", "dll.symbol"), RTM_ERR);
	cout << "initfunc" << endl;

    // Call component initialize function
    if (init_func(this))
      {
		cout << "init func executed" << endl;
		return RTM_OK;
      }
	
	cout << "init func executed failed" << endl;
    return RTM_ERR;
  }
  
  
  /*!------------------------------------------------------------
   * [CORBA interface] unload module
   * @brief Unload component shared object file
   *------------------------------------------------------------*/
  RtmRes RtcManager::unload(const char* pathname)
  {
    cout << "unload command is not implemented" << endl;
    return true;
  }
  
  
  /*!------------------------------------------------------------
   * [CORBA interface] create_component
   * @brief Instansiate a component
   *------------------------------------------------------------*/
  RTCBase_ptr RtcManager::create_component(const char* module_name,
										   CORBA::String_out component_name)
  {
	string comp_name;
	RtcBase* comp = createComponent(string(module_name), comp_name);

    component_name = CORBA::string_dup(comp_name.c_str());

	// added 2004.11.7
	//	CORBA::PolicyList policies;
	//	policies.length(0);
	//	PortableServer::POAManager_var rootManager = m_pPOA->the_POAManager(); 
	//	PortableServer::POA_var pPOA;
	//	pPOA = m_pPOA->create_POA(comp_name.c_str(),
	//							   NULL, policies);


	if (comp != NULL)
	  {
		CORBA::Object_ptr obj = m_pPOA->servant_to_reference(comp);
		m_pPOA->the_POAManager()->activate();		
		return RTCBase::_narrow(obj);
	  }
	else
	  {
		return RTCBase::_nil();
	  }
  }
  
  
  //------------------------------------------------------------
  // delete_component: Delete component instance by instance name
  //------------------------------------------------------------
  RtmRes RtcManager::delete_component(const char* instance_name)
  {
    //### delete component instance
    //### m_FactoryMap[instance_name]._delete(m_Components[instance_name]);
	
    return true;
  }
  
  
  //------------------------------------------------------------
  // component_factory_list: return component factory list
  //------------------------------------------------------------
  ComponentFactoryList* RtcManager::component_factory_list()
  {
    return NULL;
  }
  
  
  //------------------------------------------------------------
  // component_list: return component instance list
  //------------------------------------------------------------
  RTCBaseList* RtcManager::component_list()
  {
    return NULL;
  }
  
  
  //------------------------------------------------------------
  // command: Simple command interpreter
  //------------------------------------------------------------
  RtmRes RtcManager::command(const char* cmd, CORBA::String_out ret)
  {
    //  boost::reg_expression<char>
	//cmd_regex("([[:graph:]]+)[[:space:]]+([[:print:]]*?)");
    boost::regex args_regex("[[:space:]]",
							boost::regbase::normal | boost::regbase::icase);
    //  boost::match_results<const char*> match_results;
    vector<string>::iterator it;
    vector<string>::iterator it_end;
	
    vector<string> arglist;
    vector<string> retlist;
	
    cout << "Command received:" << cmd << endl;
	
    string s(cmd);
	
    if (boost::regex_split(back_inserter(arglist), s, args_regex)) {
      // boost::regex_match(s, match_results, cmd_regex) ) {
	  
	  
      //! arg0          arg1 arg2 ....
      //! command_name [arglist]......
      string command_name(arglist[0]);
      arglist.erase(arglist.begin());
	  
      it = arglist.begin();
      it_end = arglist.end();
	  
      cout << "Command: " << command_name << "\tArgs:" /*<< arglist*/ << endl;
      while (it != it_end) {
		cout << *it << endl;
		++it;
      }
	  
      //! Command invokation
      if (m_CmdMap.find(command_name) != m_CmdMap.end()) {
		bool r  = (m_CmdMap[command_name])(arglist, retlist);
		cout << "Retval: " << r << endl;
      } else {
		cout << "command not found." << endl;
      }
	  
	  
      string ret_string;
      it = retlist.begin();
      it_end = retlist.end();
	  
      while (it != it_end) {
		cout << (*it);
		ret_string = ret_string + (*it) + " ";
		++it;
      }
      cout << endl;
	  
      ret = CORBA::string_dup(ret_string.c_str());
    }
	
    return true;
  }
  
  
  
  //============================================================
  //
  // <<< Local interfaces >>>
  //
  //============================================================
  /*!------------------------------------------------------------
   * Initialize manager
   *------------------------------------------------------------*/
  void RtcManager::initManager()
  {
	
	
    createCommand("load", boost::bind(&RtcManager::loadCmd, this, _1, _2));
    createCommand("unload", boost::bind(&RtcManager::unloadCmd, this, _1, _2));
    createCommand("createcomp", boost::bind(&RtcManager::createComponentCmd, this, _1, _2));
    createCommand("lscomp", boost::bind(&RtcManager::listComponent, this, _1, _2));
    createCommand("lsmod", boost::bind(&RtcManager::listModule, this, _1, _2));
    createCommand("?", boost::bind(&RtcManager::commandListCmd, this, _1, _2));
	
  }
  
  /*!------------------------------------------------------------
   * Manager main loop
   *------------------------------------------------------------*/
  void RtcManager::runManager()
  {
	// start POA manager
	m_pPOAManager->activate();
	
	if (CORBA::is_nil(m_pORB))
	  {
		cerr << "Invalid ORB pointer" << endl;
		exit(1);
	  }
	m_pORB->run();
  }
  
  void RtcManager::runManagerNoBlocking()
  {
	open(0);
  }
  
  
  /*!------------------------------------------------------------
   * Activate RTCManager as a CORBA servant itself
   *------------------------------------------------------------*/
  bool  RtcManager::activateManager()
  {
	ACE_TRACE("RtcManager::activateManager()");

    //! Naming Object initialization
	//	cout << m_apConfig->getNameServer() << endl;
	m_apNaming->initNaming(m_pORB);

    //! Create naming context of host layer
    bool retval;
    retval = m_apNaming->createHostContext(m_apConfig->getHostname());
    if (retval == false)
	  {
		cerr << "Could not create host context on the Naming server:";
		cerr << m_apConfig->getNameServer() << endl;
		return false;
	  }
	//	cout << "Host context created" << endl;
    //! Create naming context of manager layer
    m_ManagerName = string("Manager") + m_apConfig->getPid();
    retval = m_apNaming->createManagerContext(m_ManagerName);
    if (retval == false)
	  {
		cerr << "Could not create manager context on the Naming server:";
		cerr << m_apConfig->getNameServer() << endl;
		return false;
	  }
	//	cout << "Manager context created" << endl;
    retval = m_apNaming->bindManager(m_ManagerName, this->_this());
    if (retval == false)
	  {
		cerr << "Could not bind manager object to the Naming server:";
		cerr << m_apConfig->getNameServer() << endl;
		return false;
	  }
	//	cout << "Manager bound" << endl;

	if (!initLogger())
	  {
		cerr << "error: activateManager()" << endl;
		return false;
	  }

    return true;
  }


  void RtcManager::initModuleProc(RtcModuleInitProc proc)
  {
	proc(this);
  }
	

  
  /*------------------------------------------------------------
   * Create command on the command interpreter
   *------------------------------------------------------------*/
  bool RtcManager::createCommand(string cmdName, RtcMgrCmd rtcCmd)
  {
    if (m_CmdMap.count(cmdName) > 0)
	  {
		m_CmdMap[cmdName] = rtcCmd;
	  }
	else 
	  {
		m_CmdMap[cmdName] = rtcCmd;
	  }
    return true;
  }
  
  
  /*------------------------------------------------------------
   * RegisterComponent
   *------------------------------------------------------------*/
  bool RtcManager::registerComponent(RtcModuleProfile& profile,
									 RtcNewFunc new_func,
									 RtcDeleteFunc delete_func)
  {

	RtcFactoryCXX* factory = new RtcFactoryCXX(profile, new_func, delete_func);
	
    return registerComponent(factory);
  }

  
  bool RtcManager::registerComponent(RtcFactoryBase* factory)
  {
	if (factory == NULL)
	  {
		return false;
	  }
	
	RtcModuleProfile profile;
	profile = factory->profile();

    // Module name should be unique.
    if (m_FactoryMap._map.find(profile.getImplementationId())
		!= m_FactoryMap._map.end())
	  {
		return false;
	  }
	
    m_FactoryMap._map[profile.getImplementationId()] = factory;
	
    // Get category from profile and create naming context of it
	string category;
	category = m_FactoryMap._map[profile.getImplementationId()]
	  ->profile().getCategory();
    m_apNaming->createCategoryContext(category);
    m_apNaming->createModuleContext(profile.getImplementationId(), category);
	
    //! If component type is Static, instantiate MaxInstance objects.
    if (profile.getComponentType() == STATIC)
	  {
		int i;
		for (i = 0; i < profile.getMaxInstance(); ++i)
		  {
			char* instance_name;
			create_component(profile.getImplementationId(), instance_name);
			CORBA::string_free(instance_name);
		  }
	  }
	
    return true;
	;
  }



  RtcBase* RtcManager::createComponent(const string& module_name,
				       string& comp_name)
  {
    ACE_Guard<ACE_Thread_Mutex> guard(m_FactoryMap._mutex);

    //! Instantiate a component by module name (comp_name)
    //! _new is Factory function of the "Component Object"
	if (m_FactoryMap._map.find(module_name) == m_FactoryMap._map.end())
	  {
		return NULL;
	  }
	
	RtcFactoryBase* factory = m_FactoryMap._map[module_name];

	// current number of component: factory->number() + 1
	if ((factory->number() + 1) >= factory->profile().getMaxInstance())
	  {
		return NULL;
	  }

    RtcBase* comp;
	comp = factory->create(this);
	
	// Pass the module profile to created component
	comp->initModuleProfile(factory->profile());

    // Create component instance name and set propery itself.
	comp_name = comp->setComponentName(factory->number());
	m_Components[comp_name] = comp;
	
	string category(factory->profile().getCategory());
	
    // Bind component object on the Naming Server
    // at a category that the component belongs at.
	comp->_this();
	if (comp->isLongNameEnable())
	  {
		m_apNaming->bindComponent(comp_name, module_name, category,
								  m_pPOA->servant_to_reference(comp));
	  }

	// Bind the component on alias at naming server
	if (comp->isAliasEnable())
	  {
		std::list<std::string>::iterator it, it_end;
		std::list<string> aliases = comp->getAliases();
		it     = aliases.begin();
		it_end = aliases.end();
		while (it != it_end)
		  {
			CORBA::Object_var obj = m_pPOA->servant_to_reference(comp);
			m_apNaming->bindObjectByFullPath(*it, obj);
			++it;
		  }
	  }
	m_Components[comp_name]->open(0);

    return m_Components[comp_name];
  }  
  
  RtcBase* RtcManager::createComponent(const string& module_name)
  {
    string comp_name;
    return createComponent(module_name, comp_name);
  }

  std::vector<RTCBase_ptr>
  RtcManager::findComponents(const string& comp_name)
  {
	ACE_TRACE("findComponents()");
	std::vector<RTCBase_ptr> comps;
	ObjectList objs;

	m_apNaming->findComponents(comp_name, objs);

	vector<CORBA::Object_ptr>::iterator it, it_end;
	it     = objs.begin();
	it_end = objs.end();

	for (it = objs.begin(); it != it_end; ++it)
	  {
		cout << ".";
		RTCBase_ptr tmp = RTCBase::_narrow(*it);
		if (!CORBA::is_nil(tmp))
		  {
			comps.push_back(tmp);
		  }
	  }
	return comps;
  }


  std::string RtcManager::bindInOut(RTCBase_ptr comp_in,
				    const std::string& inp_name,
				    RTCBase_ptr comp_out,
				    const std::string& outp_name,
				    RTM::SubscriptionType sub_type)
  {
	InPort_ptr inpt;
	OutPort_ptr outpt;
	
	inpt  = comp_in->get_inport(inp_name.c_str());
	if (CORBA::is_nil(inpt)) return std::string("");
	outpt = comp_out->get_outport(outp_name.c_str());
	if (CORBA::is_nil(outpt)) return std::string("");

	char* uuid;
	RTM::SubscriberProfile sub_prof;
	sub_prof.subscription_type = sub_type;

	if (outpt->subscribe(inpt, uuid, sub_prof) != RTM::RTM_OK)
	  {
		cerr << "subscription error" << endl;
		return std::string("");
	  }

	return std::string(uuid);
  }
 
  std::string RtcManager::bindInOutByName(const std::string& comp_name_in,
					  const std::string& inp_name,
					  const std::string& comp_name_out,
					  const std::string& outp_name,
					  RTM::SubscriptionType sub_type)
  {
    std::vector<RTCBase_ptr> incomps;
    std::vector<RTCBase_ptr> outcomps;
    incomps  = findComponents(comp_name_in);
    outcomps = findComponents(comp_name_out);
    
    if (incomps.size() == 0)
      {
	cerr << "component not found" << endl;
	return std::string("");
      }
    else if (incomps.size() > 1)
      {
	cerr << "two or more components found" << endl;
      }
    if (outcomps.size() == 0)
      {
	cerr << "component not found" << endl;
	return std::string("");
      }
    else if (outcomps.size() > 1)
	  {
	    cerr << "two or more components found" << endl;
	  }
    
    InPort_ptr inpt;
    OutPort_ptr outpt;
    std::string ret;
    ret = bindInOut(incomps[0], inp_name, outcomps[0], outp_name, sub_type);
    return ret;
  }
  
  
  //============================================================
  // Commands
  
  bool RtcManager::loadCmd(const vector<string>& cmd, vector<string>& retval)
  {
    if (cmd.size() != 2) {
      return false;
    }
    load(cmd[0].c_str(), cmd[1].c_str());
	
    cout << "load command" << endl;
	
    return true;
  }
  
  
  
  bool RtcManager::unloadCmd(const vector<string>& cmd, vector<string>& retval)
  {
    if (cmd.size() != 1) {
      return false;
    }
    load(cmd[0].c_str(), cmd[1].c_str());
	
    cout << "unload command" << endl;
	
    return true;
  }
  
  bool RtcManager::createComponentCmd(const vector<string>& cmd,
								   vector<string>& retval)
  {
    if (cmd.size() != 1) {
      return false;
    }
	
    char* instance_name;
    bool ret = create_component(cmd[0].c_str(), instance_name);
    retval.push_back(string(instance_name));
    cout << "Retvalue CreateComponent() " << instance_name << endl;
    return ret;
  }
  
  
  bool RtcManager::listComponent(const vector<string>& cmd, vector<string>& retval)
  {
    map<string, RtcBase*>::iterator it = m_Components.begin();
    map<string, RtcBase*>::iterator it_end = m_Components.end();
	
    while (it != it_end)
	  {
		retval.push_back(it->first);
		cout << it->first << endl;
		++it;
	  }
	
    return true;
  }
  
  
  bool RtcManager::listModule(const vector<string>& cmd, vector<string>& retval)
  {
    cout << "Entering List module" << endl;
    map<string, RtcFactoryBase*>::iterator it = m_FactoryMap._map.begin();
    map<string, RtcFactoryBase*>::iterator it_end = m_FactoryMap._map.end();
	
    while (it != it_end)
	  {
		retval.push_back(it->first);
		cout << it->first << endl;
		++it;
	  }
	
    return true;
  }
  
  
  bool RtcManager::commandListCmd(const vector<string>& cmd, vector<string>& retval)
  {
    map<string, boost::function2<bool, vector<string>&, vector<string>&> >::iterator it;
	
    for (it = m_CmdMap.begin(); it != m_CmdMap.end(); ++it)
	  {
		cout << it->first << " ";
		retval.push_back(it->first);
	  }
    cout << endl;
	
    return true;
  }
  
  
  
  //============================================================
  //
  // <<< Private methods >>
  //
  //============================================================
  /*!------------------------------------------------------------
   * Initialize ORB and store ORB's reference
   *------------------------------------------------------------*/
  bool RtcManager::initORB()
  {
    // Initialize ORB
    try
	  {
		int _argc = m_apConfig->getOrbInitArgc();
		char** _argv = m_apConfig->getOrbInitArgv();
		cerr << "_argv" << _argc << endl;
		for (int i = 0; i < _argc ; ++i)
		  {
			cerr << i << ": " << _argv[i] << endl;
			cerr << strlen(_argv[i]) << endl;
		  }
		m_pORB = CORBA::ORB_init(_argc, _argv);
	  }
    catch (...)
	  {
		cout << "error cought" << endl;
		return false;
	  }
    return true;
  }

  CORBA::ORB_var RtcManager::getORB()
  {
	return m_pORB;
  }
  
  
  /*!------------------------------------------------------------
   * Initialize POA and POA manager
   *------------------------------------------------------------*/
  bool RtcManager::initPOA()
  {
    //------------------------------------------------------------
    // Resolve root POA initial reference and get POA manager
    CORBA::Object_var obj = m_pORB->resolve_initial_references("RootPOA");
    m_pPOA = PortableServer::POA::_narrow(obj);
    if (CORBA::is_nil(m_pPOA))
	  {
		cerr << "Could not resolve RootPOA" << endl;
		return false;
	  }
    m_pPOAManager = m_pPOA->the_POAManager();
	cerr << "POA Manager created" << endl;
    return true;
  }
  
  PortableServer::POA_var RtcManager::getPOA()
  {
	return m_pPOA;
  }
  

  bool RtcManager::initLogger()
  {
	cout << m_apConfig->getLogFileName() << endl;
	std::string logfile;
	std::string errlogfile;
	logfile = m_apConfig->getLogFileName();
	errlogfile = m_apConfig->getErrorLogFileName();

	// Initialize Log stream
	{
	  ACE_Guard<ACE_Thread_Mutex> guard_curr(m_Log._mutex);


	  if (logfile.size() < 1)
		{
		  logfile = "./rtm" + m_apConfig->getPid() + ".log";
		}
	  m_Log._stream.open(logfile.c_str());

	  if (!m_Log._stream)
		{
		  cerr << "Local log file open error!!!" << endl;
		  return false;
		}
	}
	writeLog(0, "manager : OpenRTM-aist-0.2.0.");
	writeLog(0, "manager : Copyright (c) 2003-2004 OpenRTM-aist Project.");
	writeLog(0, "manager : RtcManager starting.");
	writeLog(0, "manager : Starting local logging.");

	// Initialize ErrorLog stream
	{
	  ACE_Guard<ACE_Thread_Mutex> guard_curr(m_Log._mutex);
	  
	  if (errlogfile.size() < 1)
		{
		  errlogfile = "./rtm" + m_apConfig->getPid() + "_err.log";
		}

	  m_ErrLog._stream.open(errlogfile.c_str());

	  if (!m_ErrLog._stream)
		{
		  cerr << "Local error log file open error!!!" << endl;
		  return false;
		}
	}
	writeErrLog(0, "manager : OpenRTM-aist-0.2.0.");
	writeErrLog(0, "manager : Copyright (c) 2003-2004 OpenRTM-aist Project.");
	writeErrLog(0, "manager : RtcManager starting.");
	writeErrLog(0, "manager : Starting local logging.");

	return true;
  }
  

  bool RtcManager::writeLog(int loglevel, std::string msg)
  {
	if (m_apConfig->getLogLevel() < loglevel)
	  {
		return false;
	  }
	time_t t;
	t = ACE_OS::time();
	string date(ACE_OS::ctime(&t));
	date.erase(date.find_last_of("\n"));
	
	ACE_Guard<ACE_Thread_Mutex> guard_curr(m_Log._mutex);
	return m_Log._stream << date << " " << msg << endl;
  }


  bool RtcManager::writeErrLog(int loglevel, std::string msg)
  {
	if (m_apConfig->getLogLevel() < loglevel)
	  {
		return false;
	  }		

	time_t t;
	t = ACE_OS::time();
	string date(ACE_OS::ctime(&t));
	date.erase(date.find_last_of("\n"));
	
	ACE_Guard<ACE_Thread_Mutex> guard_curr(m_ErrLog._mutex);
	return m_ErrLog._stream << date << " " << msg << endl;
  }



  bool RtcManager::findPath(string path, string &full_path)
  {
    full_path.erase();
	
#ifdef WIN32
    boost::regex path_regex("^[a-zA-Z]\\:\\\\.*");
#else
    boost::regex path_regex("^\\/.*");
#endif
    boost::smatch path_match;
	
    // Full path ?
    if (boost::regex_match(string(path), path_match, path_regex))
      {
		// Given pathname is full path name
		cout << "Full path: " << path << endl;
		full_path = path;
		if (!ACE_OS::access(path.c_str(), F_OK))
		  {
			full_path = path;
			return true;
		  }
		return false;
      }
    else
      { // Find the file in default path list
		list<string>::iterator it = m_apConfig->getComponentLoadPath().begin();
		list<string>::iterator it_end = m_apConfig->getComponentLoadPath().end();
		
		// Search pathname in the library path list
		while (it != it_end)
		  {
			string tmp_path((*it) + "/" + path);
			cout << "tmp_path: " << tmp_path << endl;
			if (!ACE_OS::access(tmp_path.c_str(), F_OK))
			  {
				full_path = tmp_path;
				cout << "found Full path: " << full_path << endl;
				return true;
			  }
			++it;
		  }
      }
	
    return false;
  }
  
  
  
  bool RtcManager::libnameToInitFunc(const char* libpath,
									 const char* initfunc,
									 string &initfunc_name)
  {
    if (libpath == NULL) return false;
	
    if (initfunc != NULL)
      {
		initfunc_name = initfunc;
		if (initfunc_name.size() > 0)
		  {
			return true;
		  }
		return false;
      }
	
    string libname(libpath);
    string::size_type pos = libname.find_last_of("/");
	
    if (pos == string::npos)
      {
		initfunc_name = libname + "Init";
      }
  }
  
  
}; // end of namespace RTM
