// -*- C++ -*-
/*!
 * @file RtcSubscriber.cpp
 * @brief Subscriber active object class runs on Output port
 * @date $Date$
 *
 * Copyright (c) 2004 Noriaki Ando <n-ando@aist.go.jp>
 *          National Institute of Industrial Science (AIST), Japan. All rights reserved.
 *
 * $Id$
 */

#include <ace/ACE.h>
#include "RtcSubscriber.h"

namespace RTM
{

  using namespace std;

  //------------------------------------------------------------
  // SubscriberBase
  //------------------------------------------------------------
  SubscriberBase::SubscriberBase(OutPortBase* parent, InPort_ptr inport,
				   const SubscriptionID id, SubscriberProfile profile)
	: m_pParent(parent), m_pInPort(InPort::_duplicate(inport)),
	  m_Id(id), m_Profile(profile)
	  {
		ACE_TRACE("SubscriberBase::SubscriberBase()");
		open(0);
	  };

  SubscriberBase::~SubscriberBase()
  {
	cout << "Subscriber's destructor called!!!!!!!!!!!!" << endl;
	close(0);
	CORBA::release(m_pInPort);
  }

  int SubscriberBase::open(void *args)
  {
	activate(); return 0;
  }

  void SubscriberBase::release()
  {
	delete this;
  }


  SubscriberProfile SubscriberBase::get_profile()
  {
	return m_Profile;
  }


  InPort_ptr SubscriberBase::get_inport()
  {
	return m_pInPort;
  }


  std::string SubscriberBase::get_id()
  {
	return m_Id;
  }

  //------------------------------------------------------------
  // SubscriberOnce
  //------------------------------------------------------------
  SubscriberOnce::SubscriberOnce(OutPortBase* parent, InPort_ptr inport,
								 const SubscriptionID id,
								 SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	ACE_TRACE("SubscriberOnce::SubscriberOnce()");
  }

  // Notify data update to subscriber
  bool SubscriberOnce::update()
  {
	ACE_TRACE("SubscriberOnce::update()");

	ACE_GUARD_RETURN(ACE_Thread_Mutex, guard, m_NewData.m_Mutex, false);
	m_NewData.m_Updated = true;
	m_NewData.m_Cond.signal();
	return true;
  }

  // Subscriber active object main routine
  int SubscriberOnce::svc(void)
  {
	ACE_TRACE("SubscriberOnce::svc()");

	m_NewData.m_Mutex.acquire();
	
	// Waiting for new data updated
	while (!m_NewData.m_Updated)
	  m_NewData.m_Cond.wait();
	
	if (m_NewData.m_Updated)
	  {
		//		m_pParent->push(m_InPort);
		//		m_pParent->unsubscribe(m_InPort, m_Id);
		m_NewData.m_Updated = false;
	  }
	
	m_NewData.m_Mutex.release();
  }
  


  //------------------------------------------------------------
  // SubscriberPeriodic
  //------------------------------------------------------------
  SubscriberPeriodic::SubscriberPeriodic(OutPortBase* parent,
										 InPort_ptr inport,
										 const SubscriptionID id,
										 SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	ACE_TRACE("SubscriberPeriodic::SubscriberPeriodic()");
  }


  // Notify data update to subscriber
  bool SubscriberPeriodic::update()
  {
	ACE_TRACE("SubscriberPeriodic::update()");

	// Do nothing: Periodic subscriber object push data asynchronously to InPort
	return true;
  }


  // Subscriber active object main routine
  int SubscriberPeriodic::svc(void)
  {
	ACE_TRACE("SubscriberPeriodic::svc()");
#ifdef ARTLinux
	//	art_enter();
#endif // ARTLinux
	while (1)
	  {
		sleep(1);
		//		m_pParent->push(m_InPort, m_Id);
		
		// sleep() is needed
#ifdef ARTLinux
		//art_wait();
#else
		//		ACE_OS::sleep(m_WaitTime);
#endif //ARTLinux
		
	  }
  }


  //------------------------------------------------------------
  // SubscriberNew
  //------------------------------------------------------------
  SubscriberNew::SubscriberNew(OutPortBase* parent, InPort_ptr inport,
								const SubscriptionID id,
							   SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	ACE_TRACE("SubscriberNew::SubscriberNew()");
  }


  // Notify data update to subscriber
  bool SubscriberNew::update()
  {
	ACE_TRACE("SubscriberNew::update()");

	ACE_GUARD_RETURN(ACE_Thread_Mutex, guard, m_NewData.m_Mutex, false);
	m_NewData.m_Updated = true;
	m_NewData.m_Cond.signal();
	return true;
  }


  int SubscriberNew::svc(void)
  {
	ACE_TRACE("SubscriberNew::svc()");

	while (1)
	  {
		m_NewData.m_Mutex.acquire();
		
		// Waiting for new data updated
		while (!m_NewData.m_Updated)
		  m_NewData.m_Cond.wait();
		
		if (m_NewData.m_Updated)
		  {
			if (m_pParent->push(m_pInPort, m_Id) != RTM_OK) break; 

			m_NewData.m_Updated = false;
		  }
		
		m_NewData.m_Mutex.release();	
	  }
  }


  //------------------------------------------------------------
  // SubscriberTriggerd
  //------------------------------------------------------------
  SubscriberTriggerd::SubscriberTriggerd(OutPortBase* parent,
										 InPort_ptr inport,
										 const SubscriptionID id,
										 SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	ACE_TRACE("SubscriberTriggerd::SubscriberTriggerd()");

  }


  //------------------------------------------------------------
  // SubscriberNewPeriodic
  //------------------------------------------------------------
  SubscriberNewPeriodic::SubscriberNewPeriodic(OutPortBase* parent,
											   InPort_ptr inport,
											   const SubscriptionID id,
											   SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	ACE_TRACE("SubscriberNewPeriodic::SubscriberNewPeriodic()");
  }


  //------------------------------------------------------------
  // SubscriberPeriodicNew
  //------------------------------------------------------------
  SubscriberPeriodicNew::SubscriberPeriodicNew(OutPortBase* parent,
											   InPort_ptr inport,
											   const SubscriptionID id,
											   SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	ACE_TRACE("SubscriberPeriodicNew::SubscriberPeriodicNew()");

  }



  //------------------------------------------------------------
  // SubscriberPeriodicTriggerd
  //------------------------------------------------------------
  SubscriberPeriodicTriggerd::
  SubscriberPeriodicTriggerd(OutPortBase* parent,
							 InPort_ptr inport,
							 const SubscriptionID id,
							 SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	ACE_TRACE("SubscriberPeriodicTriggerd::SubscriberPeriodicTriggerd()");

  }



  //------------------------------------------------------------
  // SubscriberTriggerdPeriodic
  //------------------------------------------------------------

  SubscriberTriggerdPeriodic::
  SubscriberTriggerdPeriodic(OutPortBase* parent,
							 InPort_ptr inport,
							 const SubscriptionID id,
							 SubscriberProfile profile)
	:SubscriberBase(parent, inport, id, profile)
  {
	ACE_TRACE("SubscriberTriggerdPeriodic::SubscriberTriggerdPeriodic()");
  }




}; //end of namespace RTM






